#include <iostream>
#include <memory>
#include <cctype>
#include <limits.h>
#include <cassert>
#include <functional>
#include <map>
#include <typeinfo>

#ifdef WRITE_RULES
#define WRITE(x) cerr<<x
#else
#define WRITE(x) 
#endif

#define __ALGEBRA_IMPLEMENTATION__
#define __ALGEBRA_CPP__
#include "algebra.h"

#ifdef USE_NAMESPACE
namespace Algebra {
#endif
#if 0
} // don't indent namespace!
#endif

Expr AlgebraPi=Expr(new PiConst);
Expr AlgebraZero=Expr(0);
Expr AlgebraOne=Expr(1);

int Expr::count=0;

unsigned int Expression::mask_used=Expression::mask_used_init;

unsigned int Expression::get_bit() {
  unsigned int i;
  for(i=1;i!=0 && (i&mask_used)!=0;i=i<<1);
  if (i==0) throw runtime_error("Expression: too many mask bits requested!");
  mask_used=mask_used|i;
  //  cerr<<"(given bit "<<i<<")\n";
  return i;
}


int nbits(int x) { // numero di bit (logaritmo in base 2)
  int n;
  if (x<0) x=-x;
  for (n=0;x>0;n++,x=x>>1);
  return n;
}

const int BITS=nbits(INT_MAX); // di solito 31

inline int safe_sum(int x, int y) {
  if (nbits(x)>BITS-1 || nbits(y)>BITS-1) 
    throw overflow_error("in safe_sum");
  return x+y;
}

inline int safe_product(int x, int y) {
  if (nbits(x)+nbits(y)>BITS) throw overflow_error("in safe_product");
  return x*y;
}
  
  Expr Integer::operator+(const Integer &y) const {
    return Expr(new Integer(safe_sum(store,y.store)));
  };
  Expr Integer::operator-(const Integer &y) const {
    return Expr(new Integer(safe_sum(store,-y.store)));
  };
  Expr Integer::operator*(const Integer &y) const {
    return Expr(new Integer(safe_product(store,y.store)));
  };

  Expr Expr::derivative(const Expr &x) const {
    if (const Variable *p=dynamic_cast<const Variable *>
	(&(*x))) return (*this)->derivative(p->name());
    else throw runtime_error("Not a variable in derivative");
  }
  
  Expr& Expr::operator+=(const Expr& ee) {
    reset(new Sum(*this,ee));
    return *this;
  };
  
  Expr& Expr::operator-=(const Expr& ee) {
    reset(new Difference(*this,ee));
    return *this;
  };
  
  Expr& Expr::operator*=(const Expr& ee) {
    reset(new Product(*this,ee));
    return *this;
  };

Expr& Expr::operator/=(const Expr& ee) {
  reset(new Division(*this,ee));
  return *this;
};

Expr pow(const Expr &x, const Expr &y) {
  Expr z(new Power(x,y));
  return z;
}
  
  Expr sin(const Expr &x){return Expr(new Sin(x));};
  Expr cos(const Expr &x){return Expr(new Cos(x));};
  Expr log(const Expr &x){return Expr(new Log(x));};

  Expr MakeSystem(const Expr &x, const Expr &y) {
    return Expr(new System(x,y));
  }

Expr Equation(const Expr &x, const Expr &y) {
  return Expr(new Equality(x,y));
}

Expr operator+(const Expr &x, const Expr &y) {
  return Expr(new Sum(x,y));
}

Expr operator-(const Expr &x, const Expr &y) {
  return Expr(new Difference(x,y));
}

Expr operator*(const Expr &x, const Expr &y) {
  return Expr(new Product(x,y));
}

Expr operator/(const Expr &x, const Expr &y) {
  return Expr(new Division(x,y));
}

ostream& Binary::out(ostream &o) const {
  if (left()->precedence() < precedence()) o<<"("<<left()<<")";
  else o<<left();
  o<<name();
  if (right()->precedence() <= precedence()) o<<"("<<right()<<")";
  else o<<right();
  return o;
};

void print(const Expr &e) {
  cout<<"="<<e<<"\n";
}

// simplifications

bool less_complex(const Expr &e, const Expr &f) {
  if (typeid(*e)==typeid(*f)) {
    if (const Binary *be=dynamic_cast<const Binary*>(&*e)) {
      const Binary* bf=static_cast<const Binary*>(&*f);
      if (less_complex(be->right(),bf->right())) return true;
      if (less_complex(bf->right(),be->right())) return false;
      return less_complex(be->left(),bf->left());
    }
    if (const Unary *be=dynamic_cast<const Unary*>(&*e)) {
      const Unary* bf=static_cast<const Unary*>(&*f);
      return less_complex(be->arg(),bf->arg());
    }
    if (const Variable *be=dynamic_cast<const Variable*>(&*e)) {
      const Variable* bf=static_cast<const Variable*>(&*f);
      return be->name()<bf->name();
    }
    return &*e < &*f;
  }
  if (e->constant() && ! f->constant()) return true;
  if (!e->constant() && f->constant()) return false;
  if (dynamic_cast<const Integer*>(&*e)!=0) return true;
  return e->complexity()< f->complexity();
}

/*
class Depth {
  static int counter;
public:
  Depth() {++counter;};
  ~Depth() {--counter;};
  operator int() {return counter;};
};

int Depth::counter=0;
*/

bool Binary::simplify(Expr &result) const {
  //  Depth dep;
  if ((typeid(*left())==typeid(Double) && right()->constant()) ||
      (typeid(*right())==typeid(Double) && left()->constant())) {
    WRITE("binary: "<<Expr(this)<<" -> ");
    result.reset(new Double(eval()));
    WRITE(result<<"\n");
    return true;
  }
  Expr l=left(),r=right();
  if (l->simplify(l) || r->simplify(r)) {
    result=construct(l,r);
    return true;
  }
  if (commutative()) {
    if (typeid(*r)==typeid(*this)) {
      const Binary *bin=static_cast<const Binary *>(&*r);
      WRITE("binary: "<<Expr(this));
      result=construct(construct(l,bin->left()),bin->right());
      WRITE(" -> "<<result<<"\n");
      return true;
    }
    if (typeid(*l)==typeid(*this)) {
      const Binary *bin=static_cast<const Binary *>(&*l);
      if (less_complex(r,bin->right())) {
	WRITE("binary: "<<Expr(this));
	result=construct(construct(bin->left(),r),bin->right());
	WRITE(" -> "<<result<<"\n");
	return true;
      }
    } else {
      if (less_complex(r,l)) {
	WRITE("binary: "<<Expr(this));
	result=construct(r,l);
	WRITE(" -> "<<result<<"\n");
	return true;
      }
    }
  }
  return Expression::simplify(result);
}

bool Unary::simplify(Expr &result) const {
  if (typeid(*arg()) == typeid(Double)) {
    WRITE("unary: "<<Expr(this));
    result.reset(new Double(eval()));
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  Expr l=arg();
  if (l->simplify(l)) {
    result=construct(l);
    return true;
  }
  return Expression::simplify(result);
}

bool Sum::simplify(Expr &result) const {
  if ((mask & mask_simplified)!=0) {result.reset(this);return false;}
  if (left()->zero()) {
    WRITE("suml: "<<Expr(this));
    result=right();
    WRITE(" -> "<<result<<"\n");
    return true;
  } 
  if (right()->zero()) {
    WRITE("sum: "<<Expr(this));
    result=left();
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  if (typeid(*left())==typeid(Integer) && 
      typeid(*right())==typeid(Integer)) {
    WRITE("sum: "<<Expr(this));
    result=Expr(static_cast<const Integer&>(*left()) 
	 + static_cast<const Integer&>(*right()));
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  return Binary::simplify(result);
}

bool Product::simplify(Expr &result) const {
  if ((mask & mask_simplified)!=0) {result.reset(this);return false;}
  if ((mask & mask_simplified)!=0) return false;
  if (left()->zero() || right()->zero()) {
    WRITE("product: "<<Expr(this));
    result=AlgebraZero;
    WRITE(" -> "<<result<<"\n");
    return true;
  } 
  if (left()->one()) {
    WRITE("product: "<<Expr(this));
    result=right();
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  if (right()->one()) {
    WRITE("product: "<<Expr(this));
    result=left();
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  if (typeid(*left())==typeid(Integer) && 
      typeid(*right())==typeid(Integer)) {
    WRITE("product: "<<Expr(this));
    result=Expr(static_cast<const Integer&>(*left()) 
		* static_cast<const Integer&>(*right()));
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  return Binary::simplify(result);
}

bool Difference::simplify(Expr &result) const {
  if ((mask & mask_simplified)!=0) {result.reset(this);return false;}
  if (left()->zero()) {
    WRITE("difference: "<<Expr(this));
    result=-right();
    WRITE(" -> "<<result<<"\n");
   return true;
  } 
  if (right()->zero()) {
    WRITE("difference: "<<Expr(this));
    result=left();
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  if (typeid(*left())==typeid(Integer) && 
      typeid(*right())==typeid(Integer)) {
    WRITE("difference: "<<Expr(this));
    result=Expr(static_cast<const Integer&>(*left()) 
		- static_cast<const Integer&>(*right()));
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  return Binary::simplify(result);
}

bool Division::simplify(Expr &result) const {
  if ((mask & mask_simplified)!=0) {result.reset(this);return false;}
  if (left()->zero()) {
    WRITE("division: "<<Expr(this));
    result=AlgebraZero;
    WRITE(" -> "<<result<<"\n");
    return true;
  } 
  if (right()->one()) {
    WRITE("division: "<<Expr(this));
    result=left();
    WRITE(" -> "<<result<<"\n");
    return true;
  }
  return Binary::simplify(result);
}


// rules


bool Match::matches(const Expr &r, const Expr &pattern) {
  if (const Variable* p=dynamic_cast<const Variable *>(&*pattern)) {
    M::const_iterator i=m.find(p->name());
    if (i!=m.end()) {
      return r == i->second;
    } else {
      m[p->name()]=r;
      return true;
    }
  } else if (typeid(*r)!=typeid(*pattern)) {
    return false;
  } else if (const Binary *p=
	     dynamic_cast<const Binary *>(&*pattern)) {
    assert(dynamic_cast<const Binary *>(&*r)!=0);
    const Binary *rr=static_cast<const Binary *>(&*r);
    return matches (rr->left(), p->left())
      && matches (rr->right(), p->right());
  } else if (const Unary *p=
	     dynamic_cast<const Unary *>(&*pattern)) {
    assert(dynamic_cast<const Unary*>(&*r)!=0); 
    const Unary *rr=static_cast<const Unary *>(&*r);
    return matches(rr->arg(),p->arg());
  } //else  
  return r == pattern;
}

Expr Match::replace(const Expr &pattern) const {
  //  cout<<"replace("<<pattern<<")\n";
  //  cout<<"map: "<<*this<<"\n";
  if (const Variable* p=dynamic_cast<const Variable *>(&*pattern)) {
    //    cout<<"return "<<m[p]<<"\n";
    assert(m.find(p->name())!=m.end());
    return m.find(p->name())->second;
  }
  if (const Binary* p=dynamic_cast<const Binary *>(&*pattern)) {
    return Expr(p->construct(replace(p->left()),replace(p->right())));
  } else if (const Unary* p=dynamic_cast<const Unary *>
	     (&*pattern)) {
    return Expr(p->construct(replace(p->arg())));
  }
  return pattern;
}

ostream& Match::out(ostream& out) const {
  for (M::const_iterator p=m.begin();
       p!=m.end();++p)
    out<<(p->first)<<" -> "<<p->second<<", ";
  return out;
}

bool Rule::Apply(Expr &e) const {
  Match m(e,myfind);
  if (m) {
    WRITE("rule: "<<e<<" -> ");
    e=m.replace(myreplace);
    WRITE(e<<"\n");
    return true;
  }
  return false;
};

bool SimplifyRule::Apply(Expr &e) const {
  Match m(e,myfind);
  if (m) {
    Expr r=m.replace(myreplace);
    Expr s=r;
    r->simplify(r);
    if (less_complex(r,s)) {
      WRITE("simplifyrule: "<<e<<" -> "<<r<<"\n");
      e=r;
      return true;
    }
  }
  return false;
};

bool Ruleset::simplify(Expr &e) const {
  bool something_done=false;
  if ((e->mask & mask)!=0) return something_done;
  e.simplify();
  while (true) {
    bool done=false;
    if (const Binary* node=dynamic_cast<const Binary*>(&*e)) {
      /*
      if (node->commutative() && typeid(*e) == typeid(*(node->left()))) {
	assert(dynamic_cast<const Binary*>(*(node->left()))!=0);
	const Binary* subnode=static_cast<const Binary*>(*(node->left()));
	Expr l=subnode->right(), r=node->right();
	if (simplify(l)) done=true;
	if (simplify(r)) done=true;
	if (
      }
      */
      Expr l=node->left(), r=node->right();
      if (simplify(l)) done=true;
      if (simplify(r)) done=true;
      if (done) {
	e=node->construct(l,r);
	e.simplify();
      }
    } else if (const Unary* node=dynamic_cast<const Unary*>(&*e)) {
      Expr l=node->arg();
      if (simplify(l)) done=true;
      if (done) {
	e=node->construct(l);
	e.simplify();
      }
    }
    for (const_iterator i=begin();i!=end();++i) {
      if ((*i)->Apply(e)) {
	e.simplify();
	done=true;
	break;
      }
    }
    if (done==false) break;
    something_done=true;
  }
  e->mask|= mask;
  return something_done;
}

Ruleset::Ruleset() {
  mask=Expression::get_bit();
}

#if 0
{ // match no indent namespace
#endif
#ifdef USE_NAMESPACE
}
#endif
