#ifndef __ALGEBRA_H__
#define __ALGEBRA_H__

#include <string>
#include <cctype>
#include <vector>
#include <map>
#include <set>
#include <stdexcept>
#include <memory>
#include <cmath>
#include <strstream>


// define INLINE to get "inline" functions...
//#define INLINE
/*
  statistics:    execution   compilation
  -DINLINE -g  : 1.20 sec     ~     min
  -DINLINE -O3 : 0.63 sec     ~3.09 min
           -g  : 1.21 sec     ~1.00 min
           -O3 : 0.67 sec     ~2.32 min
 */


#  ifdef INLINE
#    define _INLINE inline
#  else
#    define _INLINE 
#  endif

#  ifdef MUTABLE
#    define _MUTABLE mutable
#  else
#    define _MUTABLE
#  endif

using namespace std;

#ifdef USE_NAMESPACE
namespace Algebra {
#endif
#if 0
}
#endif

/* gestione algebrica di alberi di espressioni */

class BadName: public runtime_error {
 public:
  BadName(const string& s): runtime_error("Bad variable name: "+s) {};
};

class ParseError: public runtime_error {
 public:
  ParseError(const string& s): runtime_error("Parse error: "+s) {};
};

class CannotResolve: public runtime_error {
 public:
  CannotResolve(const string& s): runtime_error("Cannot resolve: "+s) {};
};

class NotImplemented: public runtime_error {
 public:
  NotImplemented(const string &s): runtime_error("Not implemented: "+s) {};
};

class Expr;
class Expression;
class Integer;

extern Expr AlgebraPi;
extern Expr AlgebraZero;
extern Expr AlgebraOne;

// oggetto concreto che incapsula una espressione
class Expr { 
 public:
  _INLINE Expr();
  _INLINE Expr(int n);
  _INLINE Expr(int num,int den);
  _INLINE Expr(string s);
  _INLINE Expr(const Expr& x);
  _INLINE Expr(double x);
  _INLINE Expr(const Integer& x);         
  // da usarsi solo con puntatori nuovi-nuovi                              
  _INLINE Expr(const Expression *ptr);
  _INLINE ~Expr();
  _INLINE Expr& operator=(const Expr &ee);
  
  Expr& operator+=(const Expr& ee);
  Expr& operator*=(const Expr& ee);
  Expr& operator/=(const Expr& ee);
  Expr& operator-=(const Expr& ee);
  // Sono identiche? Es: !(x+y == y+x).
  _INLINE friend bool operator==(const Expr &x, const Expr &y);

  _INLINE Expr operator-() const;
  _INLINE void swap(Expr &x);
  _INLINE ostream& out(ostream& o) const;
  Expr derivative(const Expr &x) const;
  _INLINE double eval() const;
  set<string> var_list() const;
  _INLINE Expr replace(const map<string,Expr> &map) const;
  string var_name() const;
  _INLINE bool constant() const;
  
  _INLINE const Expression *operator-> () const {return store;};
  _INLINE const Expression &operator*() const {return *store;};
  
  _INLINE void reset(const Expression *e);
  
  _INLINE bool simplify();

  static int count;
 private:
  // e' mutable in quanto puo' cambiare l'albero di rappresentazione
  // l'unica funzione che sfrutta il mutable e' same.
  const Expression* _MUTABLE store;
};

  
class Match {
  public:
  Match(const Expr &e, const Expr& pattern) {
    bool_val=matches(e,pattern);};
  operator bool() const {return bool_val;};
  Expr operator[](string s) {return m[s];};  
  Expr operator[](const char *s) {return m[s];};
  Expr replace(const Expr& pattern) const;
  ostream &out(ostream& out) const;
private: 
  bool matches(const Expr &e, const Expr& pattern);
  bool bool_val;
  typedef map<string,Expr> M;
  M m; // pattern -> value
};

class Rule {
 protected:
  Expr myfind, myreplace;
 public:
  virtual bool Apply(Expr &e) const;
  Rule(const Expr &find, const Expr &replace):
    myfind(find), myreplace(replace) {};
  virtual ~Rule(){};
};

class SimplifyRule: public Rule {
 protected:
  Expr mysimplify;
 public:
  virtual bool Apply(Expr &e) const;
  SimplifyRule(const Expr &find, const Expr &replace):
    Rule(find,replace){};
};

class Ruleset: public vector<Rule*> {
 protected:
  unsigned int mask;
 public:
  Ruleset();
  bool simplify(Expr &e) const;
};

_INLINE Expr sqrt(const Expr &x);
Expr MakeSystem(const Expr &x, const Expr &y);
Expr Equation(const Expr &x, const Expr &y);
Expr operator+(const Expr &x, const Expr &y);
Expr operator-(const Expr &x, const Expr &y);
Expr operator*(const Expr &x, const Expr &y);
Expr operator/(const Expr &x, const Expr &y);
Expr pow(const Expr &x, const Expr &y);
Expr sin(const Expr &x);
Expr cos(const Expr &x);
Expr log(const Expr &x);
_INLINE int sign(const Expr &x);
_INLINE ostream& operator<<(ostream &out,const Expr &x);

#if 0
{
#endif
#ifdef USE_NAMESPACE
}
#endif

#  if defined (INLINE) || defined (__ALGEBRA_IMPLEMENTATION__)
#    include "algebra_impl.h"
#  endif

#endif

