#ifndef ALGEBRA_IMPL_H
#define ALGEBRA_IMPL_H

#include <typeinfo>

#ifdef USE_NAMESPACE
namespace Algebra {
#endif  
#if 0
} //don't indent namespace
#endif 

class Variable;

ostream& operator<<(ostream&out, const Expression &);

// le Expression sono un albero che rappresenta espressioni algebriche.
// le Expression sono oggetti polimorfi.

// I puntatori a Expression vengono rimpiazzati con Expr. 
// Expression.ref_count conta il numero di Expr che puntano questa Expression.
// Expression.ref_count viene aggiornato da Expr NON da Expression. 
// le Expression non vanno modificate! (tutti i metodi sono const)

// non posso copiare le Expression ma posso copiare gli Expr che ci puntano

// Le Expr sono oggetti concreti (puntatori ad Expression).

class Expression {
private:
  // numero di Expr che puntano a questo oggetto
  // questa variabile la gestisce Expr!
  mutable int ref_count; 
  friend class Expr;

public:

  mutable unsigned int mask; 
  
  static const unsigned int mask_cachedconstant=1<<0;
  static const unsigned int mask_constant      =1<<1;
  static const unsigned int mask_simplified  =1<<2;

  // bit gia` usati da subito
  static const unsigned int mask_used_init=
    mask_cachedconstant|mask_constant|mask_simplified;

  // bit gia` usati...
  static unsigned int mask_used;

  // ottieni un bit libero
  static unsigned int get_bit();
public:


  Expression(Expression const &) {
    cerr<<"Non copiare!\n";
    abort();
  };
  Expression():ref_count(0), mask(0) {
    //    cerr<<"expression() "<<this<<"\n";
  };

  virtual ~Expression(){
    //    cerr<<"~expression "<<this<<"\n";
    assert(ref_count==0);}; 

  void notImpl(string what) const {
    throw NotImplemented(string(typeid(*this).name())+"::"+what);}

  // stampa l'espressione su ostream
  virtual ostream& out(ostream& o) const {notImpl("out");return o;};

  // per determinare la precedenza degli operatori
  // precedence(+) < precedence(*)
  virtual int precedence() const {notImpl("precedence");return 0;};
  // l'espressione non contiene variabili
  virtual bool constant() const {notImpl("constant");return false;};
  
  // sono due espressioni identiche?
  // Nota che x+y != y+x in questo senso.
  bool operator==(const Expression &x) const {
    if (typeid(x)!=typeid(*this)) return false;
    return same(&x);};
  
  // come operator== ma assume che typeid(*x)==typeid(*this)
  virtual bool same(const Expression *x) const {
    notImpl("same");return false;};

  // cerca di determinare il segno. Nel caso tornino false il segno
  // non e' quello richiesto OPPURE e' incerto
  // se tornano true il segno e' SICURAMENTE quello richiesto

  virtual bool nonnegative() const {return false;};
  virtual bool nonpositive() const {return false;};
  virtual bool zero() const {return false;};
  virtual bool one() const {return false;};
  virtual bool positive() const {return false;};
  virtual bool negative() const {return false;};

  // converte (se possibile) in double
  virtual double eval() const {
    throw runtime_error(string("cannot eval ")+typeid(*this).name());};
  
  // calcola la derivata rispetto alla variabile x
  virtual Expr derivative(const string &x) const {
    notImpl("derivative");return AlgebraZero;};

  // crea una nuova Espressione in cui ho rimpiazzato le variabili
  // indicate in map con le rispettive Expr
  virtual Expr replace(const map<string,Expr> &map) const {
    notImpl("replace");return AlgebraZero;};
  
  // aggiunge a list le variabili libere che compaiono in questa espressione
  virtual void var_list(set<string> &list) const {};

  virtual int complexity() const {return 1;};

  // cerca di semplificare, se fa qualcosa torna true altrimenti false
  // in ogni caso memorizza il risultato in result
  // se torna true conviene riprovare a semplificare ancora.
  virtual bool simplify(Expr &result) const {
    mask|=mask_simplified;
    result.reset(this);
    return false;
  };
};



inline ostream &operator<<(ostream &out,const Expression &x) {
  return x.out(out);
};

class Integer;

//espressioni terminali
class Leaf: public Expression {
private:
public:
  Leaf() {};
};

//espressioni terminali costanti
class Constant: public Leaf {
private:
public:
  virtual bool constant() const {return true;};
  virtual int precedence() const {return 100;};
  virtual Expr derivative(const string &x) const {return AlgebraZero;};
  virtual Expr replace(const map<string,Expr> &map) const {
    return Expr(this);};
};

// tipi numerici

// Double e' un numero rappresentato da un double. 
// Un Double rappresenta un numero fissato ma noto approssimativamente.
// Si SUPPONE (cosa falsa!) che l'errore tra il Double e il vero 
// numero che rappresenta
// sia sempre minore di Double::eps. NON si suppone altro.
// Di conseguenza i Double NON possono essere convertiti in interi
// (che sono tipi conosciuti esattamente).
// E' invece naturale convertire interi in Double quando si
// hanno operandi misti.

class Double: public Constant {
 private:
  double x;
  // mettere eps=0 PUO' essere utile
  static const double eps=0.0; //1E-5;
 public:
  Double(double xx): x(xx) {};
  virtual bool same(const Expression *p) const {
    return (static_cast<const Double *>(p))->x==x;};
  virtual ostream& out(ostream &o) const {return o.form("%.3e",this->eval());};
  //  virtual Expression *duplicate() const {return new Double(*this);};
  
  virtual bool nonnegative() const {return x>=eps;};
  virtual bool positive() const {return x>eps;};
  virtual bool nonpositive() const {return x<=-eps;};
  virtual bool negative() const {return x<-eps;};
  virtual bool zero() const {return eps==0.0 && x==0.0;};
  virtual bool one() const {return eps==0.0 && x==1.0;};
  
  virtual double eval() const {return x;};
};

class Integer: public Constant {
private:
public:
  Integer(int n): store(n){};
  // il costruttore-copia e' quello di default
  // e anche la copia.
  virtual bool same(const Expression *p) const {
    return (static_cast<const Integer *>(p))->store==store;};
  virtual ostream& out(ostream &o) const {return o<<store;};
  virtual int precedence() const {return /*sign()>=0 ?*/ 20 /* : 0*/;};
  virtual double eval() const {return store;};
  bool zero() const {return store==0;};
  bool one() const {return store==1;};

  Expr Integer::operator+(const Integer &y) const;
  Expr Integer::operator-(const Integer &y) const;
  Expr Integer::operator*(const Integer &y) const;
  
 private:
  int store;
};

class PiConst: public Constant {
public:
  PiConst() {};
  virtual bool same(const Expression *p) const {return true;};
  virtual ostream& out(ostream &o) const {return o<<name();};
  virtual string name() const {return "Pi";};
  virtual bool positive() const {return true;};
  virtual bool nonnegative() const {return true;};
  virtual double eval() const {return M_PI;};
};

class Variable: public Leaf {
private:
  friend class Match;
  const string my_name;
public:
  bool same(const Expression *p) const {
    return (static_cast<const Variable *>(p))->name()==name();};

  const string &name() const {return my_name;};
  
  explicit Variable(const string n): my_name(n) {
    //if (!isalpha(name()[0])) throw BadName(name());
  };
  // il costruttore-copia e' quello di default
  
  virtual ostream& out(ostream &o) const {return o<<name();};
  //  virtual Expression* duplicate() const {return new Variable(*this);};

  virtual int precedence() const {return 100;};

  virtual Expr derivative(const string &x) const {
    if (x==name()) return AlgebraOne;
    else return AlgebraZero;};

  virtual bool constant() const {return false;};

  virtual Expr replace(const map<string,Expr> &map) const {
    std::map<string,Expr>::const_iterator i=map.find(name());
    if (i!=map.end()) return i->second;
    else return Expr(this);}
  
  virtual void var_list(set<string> &list) const {list.insert(name());};
};

class Binary: public Expression {
public:
  virtual bool constant() const {
    if ((mask & mask_cachedconstant)!=0) return (mask&mask_constant)!=0;
    if (left()->constant() && right()->constant()) {
      mask |= (mask_constant|mask_cachedconstant);
      return true;
    } else {
      (mask |= (mask_cachedconstant)) /* &= ~mask_constant*/ ;
      return false;
    }
  };
  virtual Expr replace(const map<string,Expr> &map) const {
    Expr l=left()->replace(map);
    Expr r=right()->replace(map);
    if (l.operator->() == left().operator->() && 
	r.operator->() == right().operator->()) return Expr(this);
    else return construct(l,r);
  };
  virtual void var_list(set<string> &list) const {
    left()->var_list(list);
    right()->var_list(list);}
  
  virtual string name() const=0;
  virtual ostream& out(ostream &o) const;
  const Expr &left() const {return store_left;};
  const Expr &right() const {return store_right;};
  virtual bool commutative() const {return false;};
  virtual Expr construct(const Expr& x,const Expr &y) const =0;
  virtual bool same(const Expression *ex) const {
    return 
      (*left() ) == *((static_cast<const Binary *>(ex))->left() ) && 
      (*right()) == *((static_cast<const Binary *>(ex))->right());};
  
  virtual bool simplify(Expr &result) const;
  virtual int complexity() const {
    return 
      left()->complexity() + right()->complexity()
      + ((right()->constant() && ! left()->constant()) ? 1:0); 
  };
protected:
  const Expr store_left;
  const Expr store_right;
  Binary(const Expr &x, const Expr &y): 
    store_left(x), store_right(y) {};
};

class Unary: public Expression {
 public:
  const Expr& arg() const {return store_arg;};
  virtual Expr construct(const Expr &x) const =0;
  virtual int precedence() const {return 200;}; 
  virtual bool constant() const {
    return arg()->constant();};
  virtual Expr replace(const map<string,Expr> &map) const {
    const Expr a=arg()->replace(map);
    if (a.operator->() == arg().operator->()) return Expr(this);
    else return construct(a);};
  virtual void var_list(set<string> &list) const {
    arg()->var_list(list);};
  virtual bool same(const Expression *p) const {
    return *arg() == *(static_cast<const Unary *>(p)->arg());};
  virtual bool simplify(Expr &result) const;
  
 protected:
  Unary(const Expr &arg): store_arg(arg) {};
  Expr store_arg;
};

class System: public Binary {
public:
  System(const Expr &x, const Expr &y): Binary(x,y) {};
  virtual string name() const {return ", ";};
  virtual int precedence() const {return 6;};
  virtual bool commutative() const {return false;};
  virtual Expr construct (const Expr &x, const Expr &y) const {
    return Expr(new System(x,y));};
};

class Equality: public Binary {
public:
  Equality(const Expr &x,const Expr &y): Binary(x,y) {};
  virtual string name() const {return " = ";};
  virtual Expr construct(const Expr &x, const Expr &y) const {
    return Expr(new Equality(x,y));};
  virtual int precedence() const {return 8;};
  virtual bool commutative() const {return true;};
};

class Sum: public Binary {
 public:
  Sum(const Expr &x,const Expr &y): Binary(x,y) {};
  virtual string name() const {return " + ";};
  virtual int precedence() const {return 10;};
  virtual bool commutative() const {return true;};
  virtual Expr construct(const Expr &x, const Expr &y) const {
    return Expr(new Sum(x,y));}
  virtual double eval() const {return left()->eval()+right()->eval();};
  virtual Expr derivative(const string &x) const {
    return left()->derivative(x) + right()->derivative(x);};
  virtual bool simplify(Expr &result) const;
};

class Difference: public Binary {
 public:
  Difference(const Expr &x,const Expr &y): Binary(x,y) {};
  virtual string name() const {return " - ";};
  virtual int precedence() const {return 10;};
  virtual bool commutative() const {return false;};
  virtual Expr construct(const Expr &x, const Expr &y) const {
    return Expr(new Difference(x,y));}
  virtual double eval() const {return left()->eval()-right()->eval();};
  virtual Expr derivative(const string &x) const {
    return left()->derivative(x) -
      right()->derivative(x);};
  virtual bool simplify(Expr &result) const;
};

class Product: public Binary {
 public:
  Product(const Expr &x,const Expr &y): Binary(x,y) {};
  virtual string name() const {return " * ";};
  virtual int precedence() const {return 20;};
  virtual bool commutative() const {return true;};
  virtual Expr construct(const Expr &x, const Expr &y) const {
    return Expr(new Product(x,y));}
  virtual double eval() const {return left()->eval() * right()->eval();};
  virtual Expr derivative(const string &x) const {
    return left()->derivative(x) * right() +
      left()*right()->derivative(x);};
  virtual bool simplify(Expr &result) const;
};

class Division: public Binary {
 public:
  Division(const Expr &x,const Expr &y): Binary(x,y) {};
  virtual string name() const {return " / ";};
  virtual int precedence() const {return 20;};
  virtual Expr construct(const Expr &x, const Expr &y) const {
    return Expr(new Division(x,y));}
    virtual double eval() const {return left()->eval() / right()->eval();};
  virtual Expr derivative(const string &x) const {
    return (left()->derivative(x) * right() -
      left()*right()->derivative(x))/
      pow(right(),Expr(2));};
  virtual bool simplify(Expr &result) const;
};

class Power: public Binary {
 public:
  Power(const Expr &x,const Expr &y): Binary(x,y) {};
  virtual string name() const {return "^";};
  virtual int precedence() const {return 30;};
  virtual Expr construct(const Expr &x, const Expr &y) const {
    return Expr(new Power(x,y));};
  virtual double eval() const {return std::pow(left()->eval(),
					       right()->eval());};
  virtual Expr derivative(const string &x) const {
    assert(right()->constant());
    return right()*pow(left(),right()-AlgebraOne)*
      left()->derivative(x);};
};

class Opposite: public Unary {
public:
  Opposite(const Expr &x): Unary(x) {};
  virtual ostream& out(ostream &o) const {return o<<"-"<<arg();};
  virtual Expr construct(const Expr &x) const {return Expr(new Opposite(x));};
  virtual double eval() const {return -arg()->eval();};
  virtual Expr derivative(const string &x) const {
    return -arg()->derivative(x);};
};

class Function: public Unary {
public:
  virtual string name() const=0;
  virtual ostream& out(ostream &o) const {
    return o<<name()<<"("<<arg()<<")";};
  virtual int precedence() const {return 200;}; 
protected:
  Function(const Expr &arg): Unary(arg) {};
};

class Sin: public Function {
public:
  Sin(const Expr &x): Function(x) {};
  virtual string name() const {return "sin";};
  virtual Expr construct(const Expr &x) const {
    return Expr(new Sin(x));};
  virtual double eval() const {return std::sin(arg()->eval());};
  virtual Expr derivative(const string &x) const {
    return cos(arg())*arg()->derivative(x);};
};

class Cos: public Function {
public:
  Cos(const Expr &x): Function(x) {};
  virtual string name() const {return "cos";};
  virtual Expr construct(const Expr &x) const {
    return Expr(new Cos(x));};
  virtual double eval() const {return std::cos(arg()->eval());};
  virtual Expr derivative(const string &x) const {
    return -sin(arg())*arg()->derivative(x);};
};

 class Log: public Function {
 public:
   Log(const Expr &x): Function(x) {};
   virtual string name() const {return "log";};
   virtual Expr construct(const Expr &x) const {
     return Expr(new Log(x));};
   virtual double eval() const {return std::log(arg()->eval());};
   virtual Expr derivative(const string &x) const {
     return arg()->derivative(x)/arg();};
 };

// Expr inline functions

#ifdef __ALGEBRA_CPP__
#  define __INCLUDE_INLINE__
#endif
#ifdef INLINE
#  define __INCLUDE_INLINE__
#endif

#ifdef __INCLUDE_INLINE__

_INLINE Expr::Expr(): store(AlgebraZero.store) {store->ref_count++;count++;};

_INLINE Expr::Expr(int n): store(new Integer(n)) {store->ref_count++;count++;};

_INLINE Expr::Expr(int num,int den): store(new Division(Expr(num),Expr(den))) {
  store->ref_count++;count++;};

_INLINE Expr::Expr(const Integer &x): store(new Integer(x)) {
  store->ref_count++;count++;};

_INLINE Expr::Expr(string s): store(new Variable(s)) {store->ref_count++;count++;};

_INLINE Expr::Expr(const Expr& x): store(x.store) {
  store->ref_count++;count++;};

_INLINE Expr::Expr(double x): store(new Double(x)) {store->ref_count++;count++;};

_INLINE Expr::Expr(const Expression *ptr): store(ptr) {store->ref_count++;count++;};

_INLINE Expr::~Expr() {
  //  if (count%100==0)
  //    cerr<<"{expr usage: "<<count<<"}"<<char(13);
  --store->ref_count;
  if (store->ref_count==0) delete store;
  --count;};

_INLINE void Expr::reset(const Expression *e) {
  if (e!=store) {
    e->ref_count++;
    store->ref_count--;
    if (store->ref_count==0) delete store;
    store=e;
  }
};

_INLINE Expr& Expr::operator=(const Expr &ee) {
  reset(ee.store);
  return *this;
};

_INLINE void Expr::swap(Expr &x) {
  const Expression* f=store;
  store=x.store;
  x.store=f;
};

  //_INLINE ostream& Expr::out(ostream& o) const {
  //  return expression()->out(o);}

_INLINE double Expr::eval() const {return (*this)->eval();};

_INLINE Expr Expr::replace(const map<string,Expr> &map) const {
  return (*this)->replace(map);}

_INLINE set<string> Expr::var_list() const {
  set<string> set;
  (*this)->var_list(set);
  return set;
};

_INLINE string Expr::var_name() const {
  const Variable* var=dynamic_cast<const Variable *>(store);
  if (var) {
    return var->name();
  } else return "";
}

// inline Expr related functions

  _INLINE Expr sqrt(const Expr &x) {
    return pow(x,AlgebraOne/Expr(2));
  };

_INLINE Expr Expr::operator-() const {return Expr(new Opposite(*this));};

_INLINE bool operator==(const Expr &x, const Expr &y) {
  if (x.store==y.store) return true;
  if (typeid(*x.store)!=typeid(*y.store)) return false; 
#ifdef MUTABLE
  if (x.store->same(y.store)) {
    // unifica gli alberi di rappresentazione
    if (x.store->n_ref>y.store->n_ref) {
      if ((--y.store->ref_count)==0) delete y.store;
      y.store=x.store;
      x.store->ref_count++;
    } else {
      if ((--x.store->ref_count)==0) delete x.store;
      x.store=y.store;
      y.store->ref_count++;
    }
    return true;
  } //else
  return false;
#else
  return x.store->same(y.store);
#endif
}

bool less_complex(const Expr &e,const Expr &f);

_INLINE ostream& operator<<(ostream &out,const Expr &x) {
  return x->out(out);
}

_INLINE bool Expr::simplify() {
  if ((*this)->simplify(*this)) {
    while ((*this)->simplify(*this));
    return true;
  }
  return false;
}; 

_INLINE bool Expr::constant() const {return (*this)->constant();};
  
#endif

#if 0
{ // match don't indent namespace
#endif
#ifdef USE_NAMESPACE
}
#endif

#endif

