#include <map>
#include <iostream>
#include "parser.h"

#ifdef USE_NAMESPACE
using namespace Algebra;
#endif

namespace {
  enum TokenType{/*bad_token, */ eof_token, integer_token, double_token, command_token, identifier_token};

  typedef Expr (*FUN2)(const Expr &,const Expr &);
  typedef Expr (*FUN1)(const Expr &);

  struct myop {
    int precedence;
    union {
      FUN2 n2;
      FUN1 n1;
    } f;
    myop(int prec,FUN2 ff): precedence(prec) {f.n2=ff;};
    myop(int prec,FUN1 ff): precedence(prec) {f.n1=ff;};
    myop() {};
    myop(const myop &x):precedence(x.precedence), f(x.f) {};
  };

  class TokenReader {
  public:
    TokenReader(istream &i): in(i), rejected(0) {add_default();};
    static void add_command2(string s, int precedence, FUN2 fn){
      commands[s]=myop(precedence,fn);};
    static void add_command1(string s, int precedence, FUN1 fn){
      commands[s]=myop(precedence,fn);};
    static void add_default();
    TokenType get_token();
    int last_integer(){return integer;};
    string last_string(){return command;};
    double last_double(){return store_double;};
    void reject(){rejected=1;};
    int precedence() const {
      MAP::iterator f=commands.find(command);
      assert(f!=commands.end());
      return f->second.precedence;};
    FUN2 fn2() {
      MAP::iterator f=commands.find(command);
      assert(f!=commands.end());
      return f->second.f.n2;};
    FUN1 fn1() {
      MAP::iterator f=commands.find(command);
      assert(f!=commands.end());
      return f->second.f.n1;};
  private:
    istream &in;
    int rejected;
    TokenType token;
    int integer;
    double store_double;
  
    typedef map<string,myop> MAP;
    string command;
    static MAP commands;
  };
  
  TokenReader::MAP TokenReader::commands;
  
  Expr myderivative(const Expr &x, const Expr &y) {
    return x.derivative(y);
  }

  void TokenReader::add_default() {
    static int done=0;
    if (done) return;
    add_command2(",",-1,MakeSystem);
    add_command2("=",0,Equation);
    add_command2("+",1,operator+);
    add_command2("-",1,operator-);
    add_command2("*",2,operator*);
    add_command2("/",2,operator/);
    add_command2("^",3,pow);
    add_command2("%",1,myderivative);
    add_command2("(",10,0);
    add_command2(")",10,0);
    done=1;
  }

  TokenType TokenReader::get_token() {
    char c;
    if (rejected) {
      rejected=0;
      return token;
    }
    command="";
    while (isspace(c=in.get())) {
      if (c=='\n') return token=eof_token;
    } 
    if (!in || c==-1) return token=eof_token;
    if (isdigit(c)) {
      int num=0,den=1;
      bool dot=false;
      do {
	if (isdigit(c)) {
	  num=10*num+(c-'0');
	  if (dot) den*=10;
	} else if (c=='.') {
	  if (dot) break;
	  dot=true;
	} else break;
	c=in.get();
      } while (in);                                                   
      in.putback(c);
      if (dot) {
	store_double=(double)num/(double)den;
	return token=double_token;
      } else {
	integer=num;
	return token=integer_token;
      }                                                               
    } else if (isalpha(c) || c=='_' || c=='$') {
      command=string(1,c);
      while(true) {
	c=in.get();
	if (!in) break;
	if (!(isalnum(c) || c=='_' || c=='\'' || c=='$')) 
	  {in.putback(c);break;}
	command+=c;
      }
      return token=identifier_token;
    } else {
      MAP::iterator i=commands.find(string(1,c));
      if (i!=commands.end()) {
	command=i->first;
	return token=command_token;
      }
      throw ParseError("unknown token '"+string(1,c)+"'");
    }
  }

  // legge da in una espressione che contiene operatori di precedenza
  // strettamente minore a precedence.

  Expr createExpr(TokenReader &in, int precedence=-100) {
    Expr e; //espressione valutata fin'ora
    switch(in.get_token()) {
    case command_token:
      {
	string cmd=in.last_string();
	if (cmd=="(") {
	  e=createExpr(in);
	  if (in.get_token()!=command_token || in.last_string()!=")")
	    throw ParseError("')' expected");
	} else if (cmd=="-") {
	  e=-createExpr(in,2);
	} else throw ParseError("'"+cmd+"' unexpected");
      }
      break;
    case integer_token:
      e=Expr(in.last_integer());
      break;
    case double_token:
      e=Expr(in.last_double());
      break;
    case identifier_token:
      if (in.last_string() == "Pi") e=AlgebraPi;
      else if (in.last_string() == "sin") e=sin(createExpr(in,200));
      else if (in.last_string() == "cos") e=cos(createExpr(in,200));
      else if (in.last_string() == "sqrt") e=sqrt(createExpr(in,200));
      else if (in.last_string() == "log") e=log(createExpr(in,200));
      else e=Expr(in.last_string());
      break;
    case eof_token: throw ParseError("unexpected end of file");
    default: throw ParseError("unexpected token");
    }
  
    while (true) {
      switch(in.get_token()) {
      case command_token: {
	string cmd=in.last_string();
	if (cmd==")") {in.reject();return e;}
	if (in.precedence()<=precedence) {in.reject();return e;}
	FUN2 f=in.fn2();
	if (f==0) throw ParseError("unexpected token "+cmd);
	e=(*f)(e,createExpr(in,in.precedence()));
      }
      break;
      case eof_token:
	in.reject();
	return e;
      default:
	throw ParseError("internal error");
      }
    }
  }

}

istream& operator>>(istream& in, Expr &x) {
  TokenReader r(in);
  x=createExpr(r);
  return in;
}

